import functools
import io
import typing as t
from copy import deepcopy

import a2.plotting.utils_plotting
import a2.utils
import matplotlib
import matplotlib.pyplot as plt
from PIL import Image


def parallel_plot(
    plot_fn: t.Callable,
    data: t.Iterable,
    grid_shape: tuple,
    figsize: t.Optional[t.List[float]] = None,
    aspect: str = "auto",
    kwargs=None,
    height_ratios: list = None,
    axes_backend: str = "matplotlib",
    processes: int = -1,
):
    """
    Plot figure in parallel.

    Parameters:
    ----------
    plot_fn: Plotting function that will be applied to all axes
    data: Zipped data over all axes, expected shape [number of axes]
    grid_shape: Shape of axes grid making up the figure
    figsize: Size of the figure
    aspect: Aspect ratio of the axes ("auto"/"equal")
    kwargs: Key word arguments for plotting function
    height_ratios: Height ratios of axes along the rows of the figure
    axes_backend: How axes are generated ("default"/"matplotlib")
    processes: Number of plots generated in parallel

    Returns
    -------
    Figure
    """

    if axes_backend == "default":
        fig, axes, axes_colorbar = a2.plotting.utils_plotting.create_axes_grid(
            n_columns=grid_shape[1],
            n_rows=grid_shape[0],
            figure_size=figsize,
            colorbar_width=0.02,
            spacing_x=0.03,
            spacing_y=0.03,
        )
    elif axes_backend == "matplotlib":
        fig, axes = plt.subplots(
            *grid_shape, figsize=figsize, height_ratios=height_ratios, subplot_kw={"aspect": aspect}, squeeze=False
        )
    else:
        raise ValueError(f"{axes_backend=} not available")
    if kwargs is None:
        kwargs = {}
    rastered_images = a2.utils.utils.parallelize(
        functools.partial(_parallel_plot_worker, function_to_plot=plot_fn, kwargs=kwargs),
        zip(data),
        processes=processes,
    )
    for ax, rastered in zip(axes.ravel(), rastered_images):
        ax.imshow(rastered, aspect=aspect)
        ax.get_xaxis().set_ticks([])
        ax.get_yaxis().set_ticks([])
        for spine in ax.spines.values():
            spine.set_visible(False)

    plt.subplots_adjust(hspace=0, wspace=0)
    return fig


def _parallel_plot_worker(data, function_to_plot: t.Callable, kwargs):
    """Helper function for parallelization, where all plots of individual axes are saved in a buffer
    and plotted in the main function as an image (`plt.imshow`)"""
    fig = plt.figure()
    matplotlib.font_manager._get_font.cache_clear()  # necessary to reduce text corruption artifacts
    axes = plt.axes()

    function_to_plot(*data, fig=fig, ax=axes, **kwargs)
    pil_img = rasterize(fig)
    plt.close()

    return pil_img


def rasterize(fig: plt.figure):
    """Helper function for rasterization"""
    buf = io.BytesIO()
    fig.savefig(buf, format="png", bbox_inches="tight", dpi=400)
    buf.seek(0)
    pil_img = deepcopy(Image.open(buf))
    buf.close()

    return pil_img



import ast
import json
import logging
import pathlib
import typing as t

import a2.dataset
import a2.utils
import numpy as np
import pandas
import pandas as pd
import xarray


def _convert_str_to_dict(x: str) -> object:
    """
    Converts non-nan strings to dictionary/list

    Some types like dictionary are not recovered when loaded from .nc/.csv file.
    This function converts them back to original type by excluding nan-values
    as they through an exception.
    Parameters:
    ----------
    x: string to convert

    Returns
    -------
    dictionary/list/tuple if in form of string otherwise string
    """
    try:
        if pd.isnull(x) or x == "nan" or x.strip() == "":
            return np.nan
        else:
            return ast.literal_eval(x)
    except Exception as e:
        raise Exception(f"Couldn't convert {x}") from e


def convert_str_to_dict(data_array: np.ndarray, processes=-1) -> np.ndarray:
    """
    Parallelize _convert_str_to_dict
    """
    return a2.utils.utils.parallelize(
        _convert_str_to_dict,
        data_array,
        processes=processes,
        single_arg=True,
    )


def reset_index_coordinate(ds: xarray.Dataset) -> xarray.Dataset:
    """
    Resets index variable to increasing integer values starting from 0

    Parameters:
    ----------
    ds: xarray.Dataset

    Returns
    -------
    dataset with reset index variable
    """
    ds["index"] = np.arange(np.shape(ds["index"].values)[0])
    return ds


def load_tweets_dataset(
    filename: t.Union[str, pathlib.Path],
    raw: bool = False,
    reset_index_raw: bool = True,
    reset_index: bool = True,
    drop_variables: t.Optional[list[str]] = None,
    convert_bounding_box: bool = False,
    open_dataset=False,
) -> xarray.Dataset:
    """
    loads dataset from disk and converts columns into convenient data formats

    Parameters:
    ----------
    filename: Filename of cn file
    raw: Just load file, no conversions
    reset_index: Reset index coordinate
    reset_index_raw: Reset index coordinate even when `raw="true"`
    drop_variables: List of variables to drop
    convert_bounding_box: Convert bounding box coordinates from string to dictionary
    open_dataset: Open the dataset instead of loading it in memory

    Returns
    -------
    dataset of tweets
    """
    if open_dataset:
        ds = xarray.open_dataset(filename, drop_variables=drop_variables)
    else:
        ds = xarray.load_dataset(filename, drop_variables=drop_variables)
    if reset_index_raw and "index" in ds.variables:
        ds = reset_index_coordinate(ds)
    if raw:
        return ds
    if "bounding_box" in ds.variables and convert_bounding_box:
        ds["bounding_box"] = (
            ["index"],
            convert_str_to_dict(ds["bounding_box"].values),
        )
    if "created_at" in ds.variables:
        ds["created_at"] = (["index"], pd.to_datetime(ds.created_at).values)
    if "author_id" in ds.variables:
        ds["author_id"] = (["index"], ds["author_id"].astype(int).values)
    if reset_index and "index" in ds.variables:
        ds = reset_index_coordinate(ds)
    return ds


def load_tweets_dataframe_from_jsons(
    list_of_filenames: t.Sequence,
) -> pandas.DataFrame:
    """
    loads json files and converts them to single dataframe

    Parameters:
    ----------
    list_of_filenames: list of filenames of json file

    Returns
    -------
    dataframe of tweets
    """
    dataframe_list = []
    for filename in list_of_filenames:
        with open(filename) as json_file:
            for json_response in json.loads(json_file.read()):
                dataframe_list.append(pd.json_normalize(json_response["data"]))
    return pd.concat(dataframe_list, ignore_index=True)


def load_tweets_dataset_from_jsons(
    list_of_filenames: t.Sequence,
) -> xarray.Dataset:
    """
    loads json files and converts them to single xarray dataset

    Parameters:
    ----------
    list_of_filenames: list of filenames of json file

    Returns
    -------
    xarray dataset of tweets
    """
    return load_tweets_dataframe_from_jsons(list_of_filenames).to_xarray()


def _is_coordinate(ds: xarray.Dataset, key: t.Hashable) -> bool:
    return key in ds.coords.keys()


def _get_item(value):
    try:
        return value.item()
    except AttributeError:
        return value


def any_type_present(array, types_to_check):
    return any(type(_get_item(x)) in types_to_check for x in np.ndarray.flatten(array))


def save_dataset(
    ds: xarray.Dataset,
    filename: t.Union[str, pathlib.Path] = "test.nc",
    add_attributes: str = "",
    no_conversion: bool = True,
    encode_time: bool = True,
    reset_index: bool = False,
    engine: str | None = "h5netcdf",
    replace_nan_in_string_field: str = "nan",
) -> None:
    """
    saves xarray dataset to file.

    Dictionaries and lists cannot be natively saved with xarray.
    Therefore, they are converted to strings before the file is saved.
    Parameters:
    ----------
    ds: Dataset to be saved
    filename: name of file dataset is saved to
    add_attributes: Add string to attributes of `ds`
    no_conversion: Whether variables are converted
    encode_time: Whether time variable is converted
    reset_index: Whether "index" field variable is reset

    Returns
    -------
    """
    if reset_index:
        ds = reset_index_coordinate(ds.copy())
    types_to_convert = [dict, list]

    if not no_conversion:
        for k, v in ds.variables.items():
            if any_type_present(ds[k].values, types_to_check=types_to_convert):
                logging.info(f"Converting field: {k} to strings!")
                if _is_coordinate(ds, k):
                    ds = a2.dataset.utils_dataset.add_coordinates(ds, k, np.array([str(x) for x in ds[k].values]))
                else:
                    ds = a2.dataset.utils_dataset.add_variable(ds, k, np.array([str(x) for x in ds[k].values]))
    attributes = ds.attrs
    if "description" in attributes:
        attributes["description"] = attributes["description"] + add_attributes
    else:
        attributes["description"] = add_attributes
    ds.attrs = attributes
    encoding = None
    if encode_time:
        keys_time = get_time_variables(ds)
        encoding = {k: {"units": "seconds since 1900-01-01"} for k in keys_time}
    logging.info(f"... saving dataset as {filename}")
    ds.to_netcdf(filename, encoding=encoding, engine=engine)


def get_time_variables(ds):
    return [v for v in ds.variables if ds[v].dtype == "datetime64[ns]"]


def save_dataset_split(
    ds: xarray.Dataset,
    split_by: str = "day",
    key_time: str = "time",
    prefix: t.Union[str, pathlib.Path] = "ds_",
) -> None:
    """
    Splits dataset in multiple files and saves them individually.

    Splits are possible along different units of time, currently day and year.
    Parameters:
    ----------
    ds: Dataset to split
    split_by: Split by `day` or `time`
    key_time: Key name of time variable
    prefix: Prefix for new files

    Returns
    -------
    """
    if split_by == "day":
        dates, datasets = zip(*ds.resample({key_time: "1D"}))
        texts = pd.DatetimeIndex(np.array(dates)).day
    elif split_by == "year":
        dates, datasets = zip(*ds.resample({key_time: "1Y"}))
        texts = pd.DatetimeIndex(np.array(dates)).year
    else:
        raise NotImplementedError(f"split_by {split_by} not implemented yet!")
    filenames = [f"{prefix}{t}.nc" for t in texts]
    xarray.save_mfdataset(datasets, filenames)


def load_weather_stations(filename: str, drop_columns: str | list = "Unnamed: 0"):
    df = pd.read_csv(
        filename,
        dtype={"latitude": float, "longitude": float, "prcp_amt": float},
        parse_dates=["ob_end_time"],
    )
    df = df.set_index(["latitude", "longitude"])
    if drop_columns:
        df = df.drop(columns=drop_columns)
    return df


def load_multifile_dataset(files):
    """load multiple netcdf radar files converted from nimrod"""
    return xarray.open_mfdataset(files, combine_attrs="drop_conflicts")


def load_radar_dataset(files):
    """load multiple netcdf radar files converted from nimrod"""
    return load_multifile_dataset(files)



import abc

import pyproj


class UnitConverter(abc.ABC):
    @abc.abstractmethod
    def transform(self):
        pass


class ConvFromNGToLatLong(UnitConverter):
    def __init__(self):
        self.transformer = pyproj.Transformer.from_crs("epsg:27700", "epsg:4326")

    def transform(self, easting, northing, errcheck=True):
        """
        Converts units in British National Grid (easting, northing) to latitude, longitude
        Note, shift from east,north -> north,east convention
        Parameters:
        ----------
        easting: Distance from western origin in m
        northing: Distance from southern origin in m

        Returns
        -------
        latitude, longitude
        """
        return self.transformer.transform(easting, northing, errcheck=errcheck)


class ConvFromLatLongToNG(UnitConverter):
    def __init__(self):
        self.transformer = pyproj.Transformer.from_crs("epsg:4326", "epsg:27700")

    def transform(self, latitude, longitude, errcheck=True):
        """
        Converts units from latitude, longitude to British National Grid (easting, northing)
        Note, shift from north,east -> east,north convention
        Parameters:
        ----------
        easting: Distance from western origin in m
        northing: Distance from southern origin in m

        Returns
        -------
        easting, northing
        """
        return self.transformer.transform(latitude, longitude, errcheck=errcheck)



import typing as t

import a2.plotting.axes_utils
import a2.plotting.utils_plotting
import matplotlib.pyplot as plt
import numpy as np
import xarray


def set_axes_timeseries(
    ax: plt.axes,
    xlim: t.Optional[list] = None,
    ylim: t.Optional[list] = None,
    fontsize: int = 8,
    title: t.Optional[str] = None,
    label_x: t.Optional[str] = None,
    label_y: t.Optional[str] = None,
):
    """
    Customizes matplotlib axes for timeseries plotting

    Parameters:
    ----------
    ax: matplotlib axes
    xlim: Axes limits along the x-axis
    ylim: Axes limits along the y-axis
    fontsize: Size of the font
    title: Title of the axes
    label_x: Label of the x-axis
    label_y: Label of the y-axis

    Returns
    -------
    """
    if label_x is None:
        label_x = "time"
    a2.plotting.axes_utils.set_axes(
        ax, xlim=xlim, ylim=ylim, fontsize=fontsize, title=title, label_x=label_x, label_y=label_y, labelrotation_x=45
    )


def plot_timeseries(
    time: t.Union[np.ndarray, str],
    y: t.Union[np.ndarray, str],
    ds: xarray.Dataset = None,
    ax: t.Optional[plt.axes] = None,
    fig: t.Optional[plt.figure] = None,
    xlim: t.Optional[list] = None,
    ylim: t.Optional[list] = None,
    label_x: t.Optional[str] = None,
    label_y: t.Optional[str] = None,
    sort_by_time: bool = True,
    label: str = None,
    **kwargs
):
    """
    Plot timeseries with matplotlib

    Parameters:
    ----------
    time: Time values of timeseries
    y: y-values of timeseries
    ds: When provided and time/y are keys their values are drawn from `ds`
    ax: matplotlib axes
    fig: matplotlib figure
    xlim: Axes limits along the x-axis
    ylim: Axes limits along the y-axis
    label_x: Label of the x-axis
    label_y: Label of the y-axis
    sort_by_time: Sort time/y based on time before plotting
    label: Label of the plot
    kwargs: Key word arguments for plt.plot
    Returns
    -------
    """
    fig, ax = a2.plotting.utils_plotting.create_figure_axes(figure=fig, axes=ax, font_size=None)
    if ds is not None and isinstance(y, str):
        if label_y is None:
            label_y = y
        y = ds[y].values
    if ds is not None and isinstance(time, str):
        if label_x is None:
            label_x = time
        time = ds[time].values
    if sort_by_time:
        mask_time_sort = np.argsort(time)
        time = time[mask_time_sort]
        y = y[mask_time_sort]
    ax.plot(time, y, label=label, **kwargs)
    if xlim is None:
        min_x = np.nanmin(time)
        max_x = np.nanmax(time)
        xlim = [min_x, max_x]
    if ylim is None:
        min_y = np.nanmin(y)
        max_y = np.nanmax(y) * 1.1
        ylim = [min_y, max_y]
    set_axes_timeseries(ax, label_x=label_x, label_y=label_y, xlim=xlim, ylim=ylim)
    return fig



import datetime
import logging

import matplotlib.colors
import matplotlib.dates
import numpy as np
import pandas as pd

import a2.utils.constants

logger = logging.getLogger(__name__)



def _plot_colormesh(
    axes: a2.utils.constants.TYPE_MATPLOTLIB_AXES,
    xx: np.ndarray,
    yy: np.ndarray,
    matrix: np.ndarray,
    norm_object: matplotlib.colors.LogNorm | matplotlib.colors.Normalize | None,
    colormap: str,
    linewidth: float = 0,
    rasterized: bool = True,
    edgecolor: str = "face",
    convert_zeros_to_nan: bool = False,
):
    if convert_zeros_to_nan:
        matrix[matrix == 0] = np.nan
        logger.debug(f"Convert zeros to np.nan in matrix.")
    plot = axes.pcolormesh(
        xx,
        yy,
        matrix,
        norm=norm_object,
        cmap=colormap,
        linewidth=linewidth,
        rasterized=rasterized,
    )
    plot.set_edgecolor(edgecolor)
    return plot
